!******************************************************************************!
!                                                                              !
!    class_vertices.f90                                                        !
!    is part of RECOLA2 (REcursive Computation of One Loop Amplitudes)2        !
!                                                                              !
!    Copyright (C) 2016-2018 Ansgar Denner, Jean-Nicolas Lang and              !
!                            Sandro Uccirati                                   !
!                                                                              !
!    RECOLA2 is licenced under the GNU GPL version 3,                          !
!    see COPYING for details.                                                  !
!                                                                              !
!******************************************************************************!
!                                                           .---.              !
!                                                          /  .  \             !
!                                                         |\_/|   |            !
!                                                         |   |  /|            !
!          .----------------------------------------------------' |            !
!         /  .-.   class_vertices.f90                             |            !
!        |  /   \  This code was generated with REPT1L.           |            !
!        | |\_.  | Change this file and you get what you deserve. |            !
!        |\|  | /| Author: M. Chiesa, J.N. Lang                   |            !
!        | `---' | UFO date: No date                              |            !
!        |       | UFO version: 1.0                               |            !
!        |       | Generation date: 2020-07-29 19:46             /             !
!        |       |----------------------------------------------'              !
!        \       |                                                             !
!         \     /                                                              !
!          `---'                                                               !
!******************************************************************************!
module class_vertices
  use class_particles, only: n_orders
  use constants_mdl
  use dictionary, only: dict_init
  implicit none
  save

  type Branch
    integer :: lorentz_id
    integer :: perm
    integer, allocatable :: couplings(:)
    integer, dimension(n_orders) :: coupling_order
    integer :: color_id
    integer :: colorfac_id
    ! Propagator extension (only supported for massive vectors)
    ! 0: Usual propagator
    ! 1: Rxi-propagator part with pole @ M_V^2
    ! 2: Rxi-propagator part with pole @ \xi*M_V^2
    integer :: pext
  end type

  type Vertex
    type(Branch), allocatable :: branch(:)
    type(Branch), allocatable :: branchCT(:)
    type(Branch), allocatable :: branchR2(:)
  end type

  logical :: vertices_init = .false.
  logical :: vertices_filled = .false.
  integer, parameter :: n_vertices = 1204
  type(Vertex) :: vertices(0:n_vertices-1)

  ! Defines the maximum allowed number of legs participating at a vertex
  integer, parameter :: max_leg_multiplicity = 4

  integer, parameter :: FFaS0 = 1
  integer, parameter :: FFaS1 = 2
  integer, parameter :: FFaS2 = 3
  integer, parameter :: VFaF0 = 4
  integer, parameter :: VFaF1 = 5
  integer, parameter :: SVVS0 = 6
  integer, parameter :: VVS0 = 7
  integer, parameter :: VVS1 = 8
  integer, parameter :: VVS2 = 9
  integer, parameter :: VVS3 = 10
  integer, parameter :: VVS4 = 11
  integer, parameter :: VVS5 = 12
  integer, parameter :: VVS6 = 13
  integer, parameter :: VVS7 = 14
  integer, parameter :: VVS8 = 15
  integer, parameter :: VVS9 = 16
  integer, parameter :: VVS10 = 17
  integer, parameter :: UUaS0 = 18
  integer, parameter :: UUaV0 = 19
  integer, parameter :: SUUa0 = 20
  integer, parameter :: VVSS0 = 21
  integer, parameter :: UaSU0 = 22
  integer, parameter :: FFaV0 = 23
  integer, parameter :: FFaV1 = 24
  integer, parameter :: VVV0 = 25
  integer, parameter :: VVV1 = 26
  integer, parameter :: VVV2 = 27
  integer, parameter :: VVV3 = 28
  integer, parameter :: VVV4 = 29
  integer, parameter :: VVV5 = 30
  integer, parameter :: VVV6 = 31
  integer, parameter :: VVV7 = 32
  integer, parameter :: VVV8 = 33
  integer, parameter :: VVV9 = 34
  integer, parameter :: VVV10 = 35
  integer, parameter :: VVV11 = 36
  integer, parameter :: VVV12 = 37
  integer, parameter :: VVV13 = 38
  integer, parameter :: VVV14 = 39
  integer, parameter :: VVV15 = 40
  integer, parameter :: VVV16 = 41
  integer, parameter :: VVV17 = 42
  integer, parameter :: VVV18 = 43
  integer, parameter :: VVV19 = 44
  integer, parameter :: FSFa0 = 45
  integer, parameter :: FSFa1 = 46
  integer, parameter :: FSFa2 = 47
  integer, parameter :: VVVV0 = 48
  integer, parameter :: FaVF0 = 49
  integer, parameter :: FaVF1 = 50
  integer, parameter :: UaVU0 = 51
  integer, parameter :: FaF0 = 52
  integer, parameter :: FaF1 = 53
  integer, parameter :: VSVS0 = 54
  integer, parameter :: FVFa0 = 55
  integer, parameter :: FVFa1 = 56
  integer, parameter :: VUaU0 = 57
  integer, parameter :: USUa0 = 58
  integer, parameter :: VV0 = 59
  integer, parameter :: VV1 = 60
  integer, parameter :: VV2 = 61
  integer, parameter :: SVSV0 = 62
  integer, parameter :: FaFS0 = 63
  integer, parameter :: FaFS1 = 64
  integer, parameter :: FaFS2 = 65
  integer, parameter :: FaFV0 = 66
  integer, parameter :: FaFV1 = 67
  integer, parameter :: SSSS0 = 68
  integer, parameter :: SFFa0 = 69
  integer, parameter :: SFFa1 = 70
  integer, parameter :: SFFa2 = 71
  integer, parameter :: VUUa0 = 72
  integer, parameter :: SFaF0 = 73
  integer, parameter :: SFaF1 = 74
  integer, parameter :: SFaF2 = 75
  integer, parameter :: SVV0 = 76
  integer, parameter :: SVV1 = 77
  integer, parameter :: SVV2 = 78
  integer, parameter :: SVV3 = 79
  integer, parameter :: SVV4 = 80
  integer, parameter :: SVV5 = 81
  integer, parameter :: SVV6 = 82
  integer, parameter :: SVV7 = 83
  integer, parameter :: SVV8 = 84
  integer, parameter :: FaSF0 = 85
  integer, parameter :: FaSF1 = 86
  integer, parameter :: FaSF2 = 87
  integer, parameter :: SVS0 = 88
  integer, parameter :: SVS1 = 89
  integer, parameter :: FFa0 = 90
  integer, parameter :: FFa1 = 91
  integer, parameter :: VSS0 = 92
  integer, parameter :: VSS1 = 93
  integer, parameter :: VSV0 = 94
  integer, parameter :: VSV1 = 95
  integer, parameter :: VSV2 = 96
  integer, parameter :: VSV3 = 97
  integer, parameter :: VSV4 = 98
  integer, parameter :: VSV5 = 99
  integer, parameter :: VSV6 = 100
  integer, parameter :: VSV7 = 101
  integer, parameter :: SS0 = 102
  integer, parameter :: SS1 = 103
  integer, parameter :: SSVV0 = 104
  integer, parameter :: SSV0 = 105
  integer, parameter :: SSV1 = 106
  integer, parameter :: SSS0 = 107
  integer, parameter :: UVUa0 = 108
  integer, parameter :: VSSV0 = 109
  integer, parameter :: UaUS0 = 110
  integer, parameter :: VFFa0 = 111
  integer, parameter :: VFFa1 = 112
  integer, parameter :: SUaU0 = 113
  integer, parameter :: UaUV0 = 114
  complex(dp), parameter :: c1 = cmplx(sq2/2d0, 0d0, kind=dp)
  character(len=20), parameter :: c1f = "sq2/2"
  complex(dp), parameter :: c2 = cmplx(-sq2/(2d0*Nc), 0d0, kind=dp)
  character(len=20), parameter :: c2f = "-sq2/(2*Nc)"
  complex(dp), parameter :: c3 = cmplx(1d0, 0d0, kind=dp)
  character(len=20), parameter :: c3f = "1"
  complex(dp), parameter :: c4 = cmplx(0d0, sq2/2d0, kind=dp)
  character(len=20), parameter :: c4f = "i_*sq2/2"
  complex(dp), parameter :: c5 = cmplx(0d0, -sq2/2d0, kind=dp)
  character(len=20), parameter :: c5f = "-i_*sq2/2"
  complex(dp), parameter :: c6 = cmplx(-1d0/2d0, 0d0, kind=dp)
  character(len=20), parameter :: c6f = "-1/2"
  complex(dp), parameter :: c7 = cmplx(1d0/2d0, 0d0, kind=dp)
  character(len=20), parameter :: c7f = "1/2"
  complex(dp), parameter :: c8 = cmplx(-1d0/Nc, 0d0, kind=dp)
  character(len=20), parameter :: c8f = "-1/Nc"
  complex(dp), dimension(1:8) :: colorcoeffs = (/c1, c2, c3, c4, c5, c6, c7, c8&
                                               /)
  character(len=20), dimension(1:8) :: colorcoeffs_form = (/c1f, c2f, c3f, c4f &
                 , c5f, c6f, c7f, c8f/)
  integer, dimension(1:114) :: lorentz_rank = (/0, 0, 0, 1, 1, 0, 2, 3, 2, 0, 3&
                                              , 3, 2, 3, 2, 1, 3, 0, 1, 0, 0, 0&
                                              , 0, 0, 2, 1, 2, 2, 2, 2, 2, 2, 2&
                                              , 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2&
                                              , 1, 1, 1, 0, 1, 1, 1, -1, -1, 0 &
                                              , 1, 1, 1, 0, -1, -1, -1, 0, 0, 0&
                                              , 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1&
                                              , 3, 2, 0, 3, 2, 3, 3, 3, 1, 1, 1&
                                              , 1, 2, 1, -1, -1, 2, 1, 1, 0, 2 &
                                              , 2, 2, 1, 2, 1, -1, -1, 0, 2, 1 &
                                              , 0, 1, 0, 0, 1, 1, 0, 1/)
  integer, dimension(1:114) :: ff_ids = (/0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0&
                                        , 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0&
                                        , 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0&
                                        , 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0&
                                        , 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0&
                                        , 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0&
                                        , 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0&
                                        , 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0&
                                        , 0, 0, 0, 0, 0, 0, 0, 0, 0, 0/)
  integer, dimension(1:114) :: recola_base = (/5, 7, 9, 53, 58, -2, 0, 0, 0, 2 &
                                             , 0, 0, 0, 0, 0, 0, 0, 1, 161, 1  &
                                             , -4, 1, 15, 17, 0, 12, 0, 0, 0, 0&
                                             , 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 &
                                             , 0, 0, 0, 21, 27, 25, -11, 51, 57&
                                             , 166, 0, 0, -3, 35, 31, 162, 1, 0&
                                             , 0, 0, -14, 6, 10, 8, 16, 18, -1 &
                                             , 26, 23, 28, 163, 43, 48, 46, 0, &
                                             0, 13, 0, 0, 0, 0, 0, 0, 41, 47,  &
                                             45, 0, 3, 0, 0, 0, 4, 0, 14, 0, 0 &
                                             , 0, 0, 0, 0, 0, 0, -12, 0, 11, 1 &
                                             , 165, -13, 1, 33, 36, 1, 164/)
  integer, dimension(1:114) :: propagator_extension = (/0, 0, 0, 0, 0, 0, 0, 0 &
                                                      , 0, 0, 0, 0, 0, 0, 0, 0 &
                                                      , 0, 0, 0, 0, 0, 0, 0, 0 &
                                                      , 0, 0, 0, 0, 0, 0, 0, 0 &
                                                      , 0, 0, 0, 0, 0, 0, 0, 0 &
                                                      , 0, 0, 0, 0, 0, 0, 0, 0 &
                                                      , 0, 0, 0, 0, 0, 0, 0, 0 &
                                                      , 0, 0, 0, 0, 0, 0, 0, 0 &
                                                      , 0, 0, 0, 0, 0, 0, 0, 0 &
                                                      , 0, 0, 0, 0, 0, 0, 0, 0 &
                                                      , 0, 0, 0, 0, 0, 0, 0, 0 &
                                                      , 0, 0, 0, 0, 0, 0, 0, 0 &
                                                      , 0, 0, 0, 0, 0, 0, 0, 0 &
                                                      , 0, 0, 0, 0, 0, 0, 0, 0 &
                                                      , 0, 0/)
  character(len=*), parameter :: model_timestamp = "Wed Jul 29 18:07:01 2020"

  contains
    function get_timestamp_mdl() result(ret_timestamp)
      character(len=len(model_timestamp)) :: ret_timestamp
      ret_timestamp = model_timestamp
    end function get_timestamp_mdl

    function get_lorentz_rank_mdl(l_id) result(ret_lorentz_rank)
      integer, intent(in) :: l_id
      integer :: ret_lorentz_rank
      ret_lorentz_rank = lorentz_rank(l_id)
    end function get_lorentz_rank_mdl

    function get_propagator_extension_mdl(l_id) result(ret_pext)
      integer, intent(in) :: l_id
      integer :: ret_pext
      ret_pext = propagator_extension(l_id)
    end function get_propagator_extension_mdl

    function get_max_leg_multiplicity_mdl() result(ret_max_leg_multiplicity)
      integer :: ret_max_leg_multiplicity
      ret_max_leg_multiplicity = max_leg_multiplicity
    end function get_max_leg_multiplicity_mdl

    function get_fourfermion_id_mdl(l_id) result(ret_ffid)
      integer, intent(in) :: l_id
      integer :: ret_ffid
      ret_ffid = ff_ids(l_id)
    end function get_fourfermion_id_mdl

    function is_model_init_mdl() result(ret_init)
      logical :: ret_init
      ret_init = vertices_init
    end function is_model_init_mdl

    function is_model_filled_mdl() result(ret_init)
      logical :: ret_init
      ret_init = vertices_filled
    end function is_model_filled_mdl

    subroutine createCurrents
      use fill_currents, only: fill_currents_mdl
      implicit none
      call fill_currents_mdl(.false.)
    end subroutine createCurrents

    subroutine check_vertices_mdl
      use fill_currents, only: fill_currents_mdl
      implicit none
      call fill_currents_mdl(.true.)
    end subroutine check_vertices_mdl

    subroutine av(vid,bid,cid,cfid,pext,lid,perm,co,cop)
      integer, intent(in) :: vid,bid,cid,cfid,pext,lid,perm
      integer, dimension(0:), intent(in) :: co, cop
      integer scop,i
      vertices(vid)%branch(bid)%color_id = cid
      vertices(vid)%branch(bid)%colorfac_id = cfid
      vertices(vid)%branch(bid)%pext = pext
      vertices(vid)%branch(bid)%lorentz_id = lid
      vertices(vid)%branch(bid)%perm = perm
      vertices(vid)%branch(bid)%coupling_order = co
      scop = size(cop)
      allocate(vertices(vid)%branch(bid)%couplings(0:scop-1))
      do i = 0, scop-1
        vertices(vid)%branch(bid)%couplings(i) = cop(i)
      end do
    end subroutine av

    subroutine avCT(vid,bid,cid,cfid,pext,lid,perm,co,cop)
      integer, intent(in) :: vid,bid,cid,cfid,pext,lid,perm
      integer, dimension(0:), intent(in) :: co, cop
      integer scop,i
      vertices(vid)%branchCT(bid)%color_id = cid
      vertices(vid)%branchCT(bid)%colorfac_id = cfid
      vertices(vid)%branchCT(bid)%pext = pext
      vertices(vid)%branchCT(bid)%lorentz_id = lid
      vertices(vid)%branchCT(bid)%perm = perm
      vertices(vid)%branchCT(bid)%coupling_order = co
      scop = size(cop)
      allocate(vertices(vid)%branchCT(bid)%couplings(0:scop-1))
      do i = 0, scop-1
        vertices(vid)%branchCT(bid)%couplings(i) = cop(i)
      end do
    end subroutine avCT

    subroutine avR2(vid,bid,cid,cfid,pext,lid,perm,co,cop)
      integer, intent(in) :: vid,bid,cid,cfid,pext,lid,perm
      integer, dimension(0:), intent(in) :: co, cop
      integer scop,i
      vertices(vid)%branchR2(bid)%color_id = cid
      vertices(vid)%branchR2(bid)%colorfac_id = cfid
      vertices(vid)%branchR2(bid)%pext = pext
      vertices(vid)%branchR2(bid)%lorentz_id = lid
      vertices(vid)%branchR2(bid)%perm = perm
      vertices(vid)%branchR2(bid)%coupling_order = co
      scop = size(cop)
      allocate(vertices(vid)%branchR2(bid)%couplings(0:scop-1))
      do i = 0, scop-1
        vertices(vid)%branchR2(bid)%couplings(i) = cop(i)
      end do
    end subroutine avR2

    function get_colourfac_mdl(cfid)
      integer, intent(in) :: cfid
      complex(dp) :: get_colourfac_mdl
      get_colourfac_mdl = colorcoeffs(cfid)
    end function get_colourfac_mdl

    function get_colourfac_form_mdl(cfid)
      integer, intent(in) :: cfid
      character(len=20) :: get_colourfac_form_mdl
      get_colourfac_form_mdl = colorcoeffs_form(cfid)
    end function get_colourfac_form_mdl

  function get_recola_base_mdl(l_id) result(ret_rb)
      integer, intent(in) :: l_id
      integer :: ret_rb
      ret_rb = recola_base(l_id)
    end function get_recola_base_mdl

  subroutine get_vertex_branch_status_mdl(vx,xlp,branch_exists)
    integer, intent(in)  :: vx,xlp
    logical, intent(out) ::branch_exists
    select case (xlp)
    case(TreeBranch:LoopBranch)
      branch_exists = allocated(vertices(vx)%branch)
    case(CTBranch)
      branch_exists = allocated(vertices(vx)%branchCT)
    case(R2Branch)
      branch_exists = allocated(vertices(vx)%branchR2)
    end select
  end subroutine get_vertex_branch_status_mdl

  subroutine dealloc_vertex_mdl(vx,xlp)
    integer, intent(in)  :: vx,xlp
    logical ::branch_exists
    if (vx .ge. 0 .and.  vx .lt. n_vertices) then
      select case (xlp)
      case(TreeBranch:LoopBranch)
        if (allocated(vertices(vx)%branch)) deallocate(vertices(vx)%branch)
      case(CTBranch)
        if (allocated(vertices(vx)%branchCT)) deallocate(vertices(vx)%branchCT)
      case(R2Branch)
        if (allocated(vertices(vx)%branchR2)) deallocate(vertices(vx)%branchR2)
      case default
        if (allocated(vertices(vx)%branch)) deallocate(vertices(vx)%branch)
        if (allocated(vertices(vx)%branchCT)) deallocate(vertices(vx)%branchCT)
        if (allocated(vertices(vx)%branchR2)) deallocate(vertices(vx)%branchR2)
      end select
    end if
  end subroutine dealloc_vertex_mdl

  subroutine get_vertex_colour_branches_mdl(vx,xlp,cfids,nbranches)
    integer, allocatable, intent(out) :: cfids(:)
    integer, intent(out) :: nbranches
    integer, intent(in)  :: vx,xlp
    select case (xlp)
    case (TreeBranch:LoopBranch)
      allocate(cfids(size(vertices(vx)%branch)))
      cfids(:) = vertices(vx)%branch(0:)%color_id
      nbranches = size(vertices(vx)%branch)
    case(CTBranch)
      allocate(cfids(size(vertices(vx)%branchCT)))
      cfids(:) = vertices(vx)%branchCT(0:)%color_id
      nbranches = size(vertices(vx)%branchCT)
    case(R2Branch)
      allocate(cfids(size(vertices(vx)%branchR2)))
      cfids(:) = vertices(vx)%branchR2(0:)%color_id
      nbranches = size(vertices(vx)%branchR2)
    end select
  end subroutine get_vertex_colour_branches_mdl

  subroutine get_vertex_pext_mdl(vx,xlp,br,pext)
    integer, intent(in) :: vx,xlp,br
    integer, intent(out) :: pext
    select case (xlp)
    case (TreeBranch:LoopBranch)
      pext = vertices(vx)%branch(br)%pext
    case(CTBranch)
      pext = vertices(vx)%branchCT(br)%pext
    case(R2Branch)
      pext = vertices(vx)%branchR2(br)%pext
    end select
  end subroutine get_vertex_pext_mdl

  subroutine get_vertex_order_increase_mdl(vx,xlp,br,orderInc)
    integer, intent(in) :: vx,xlp,br
    integer, dimension(:), intent(out) :: orderInc
    select case (xlp)
    case (TreeBranch:LoopBranch)
      orderInc = vertices(vx)%branch(br)%coupling_order
    case(CTBranch)
      orderInc = vertices(vx)%branchCT(br)%coupling_order
    case(R2Branch)
      orderInc = vertices(vx)%branchR2(br)%coupling_order
    end select
  end subroutine get_vertex_order_increase_mdl

  subroutine get_vertex_lorentz_type_mdl(vx,xlp,br,lid)
    integer, intent(in)  :: vx,xlp,br
    integer, intent(out) :: lid
    select case (xlp)
    case (TreeBranch:LoopBranch)
      lid = vertices(vx)%branch(br)%lorentz_id
    case(CTBranch)
      lid = vertices(vx)%branchCT(br)%lorentz_id
    case(R2Branch)
      lid = vertices(vx)%branchR2(br)%lorentz_id
    end select
  end subroutine get_vertex_lorentz_type_mdl

  subroutine get_vertex_perm_mdl(vx,xlp,br,perm)
    integer, intent(in)  :: vx,xlp,br
    integer, intent(out) :: perm
    select case (xlp)
    case (TreeBranch:LoopBranch)
      perm = vertices(vx)%branch(br)%perm
    case(CTBranch)
      perm = vertices(vx)%branchCT(br)%perm
    case(R2Branch)
      perm = vertices(vx)%branchR2(br)%perm
    end select
  end subroutine get_vertex_perm_mdl

  subroutine get_vertex_colour_id_mdl(vx,xlp,br,cid)
    integer, intent(in)  :: vx,xlp,br
    integer, intent(out) :: cid
    select case (xlp)
    case (TreeBranch:LoopBranch)
      cid = vertices(vx)%branch(br)%colorfac_id
    case(CTBranch)
      cid = vertices(vx)%branchCT(br)%colorfac_id
    case(R2Branch)
      cid = vertices(vx)%branchR2(br)%colorfac_id
    end select
  end subroutine get_vertex_colour_id_mdl

  function is_zero_coupling_vertex_mdl(vx,xlp,br) result(zero)
    use class_couplings, only: get_coupling_value_mdl
    use input_mdl, only: zerocut
    integer, intent(in)  :: vx,xlp,br
    logical              :: zero
    real(dp)             :: maxvalue
    integer              :: i

    zero = .false.
    select case (xlp)
    case (TreeBranch:LoopBranch)
      maxvalue = 0d0
      do i = 1,size(vertices(vx)%branch(br)%couplings)
        maxvalue = maxvalue + &
                abs(get_coupling_value_mdl(vertices(vx)%branch(br)%couplings(i-&
                1)))
      end do
      if (maxvalue/zerocut .lt. 100d0) then
        zero = .true.
      end if
    case(CTBranch)
      maxvalue = 0d0
      do i = 1,size(vertices(vx)%branchCT(br)%couplings)
        maxvalue = maxvalue + &
                abs(get_coupling_value_mdl(vertices(vx)%branchCT(br)%couplings(&
                i-1)))
      end do
      if (maxvalue/zerocut .lt. 100d0) then
        zero = .true.
      end if
    case(R2Branch)
      maxvalue = 0d0
      do i = 1,size(vertices(vx)%branchR2(br)%couplings)
        maxvalue = maxvalue + &
                abs(get_coupling_value_mdl(vertices(vx)%branchR2(br)%couplings(&
                i-1)))
      end do
      if (maxvalue/zerocut .lt. 100d0) then
        zero = .true.
      end if
    end select
  end function is_zero_coupling_vertex_mdl

  subroutine get_vertex_ncouplings_mdl(vx,xlp,br,ncouplings)
    integer, intent(in)  :: vx,xlp,br
    integer, intent(out) :: ncouplings
    select case(xlp)
    case(TreeBranch, LoopBranch)
      ncouplings = size(vertices(vx)%branch(br)%couplings)
    case (CTBranch)
      ncouplings = size(vertices(vx)%branchCT(br)%couplings)
    case (R2Branch)
      ncouplings = size(vertices(vx)%branchR2(br)%couplings)
    end select
  end subroutine get_vertex_ncouplings_mdl

  subroutine get_vertex_coupling_ids_mdl(vx,xlp,br,cids)
    integer, intent(in)  :: vx,xlp,br
    integer, intent(out) :: cids(:)
    select case(xlp)
    case(TreeBranch, LoopBranch)
      cids(:) = vertices(vx)%branch(br)%couplings(:)
    case (CTBranch)
      cids(:) = vertices(vx)%branchCT(br)%couplings(:)
    case (R2Branch)
      cids(:) = vertices(vx)%branchR2(br)%couplings(:)
    end select
  end subroutine get_vertex_coupling_ids_mdl

  subroutine init_vertices_mdl()
    use input_mdl, only: nx, error_mdl
    use class_particles,only: n_orders,driver_timestamp
    use fill_ctparameters, only: init_CTparameters_mdl
    use fill_couplings,only: fill_couplings_mdl,fill_CTcouplings_mdl

    ! check if modelfile is compatible to drivers
    if (model_timestamp .ne. driver_timestamp) then
      call error_mdl("Model corrupted. Driver timestamp mismatch.", &
                    where='init_vertices')
    end if

    vertices_init = .true.
    call fill_couplings_mdl
    call init_CTparameters_mdl(.true.)
    call fill_CTcouplings_mdl
    call dict_init(2399)
    call createCurrents

  end subroutine init_vertices_mdl

    subroutine get_vertex_lorentz_rank_mdl(ty,cc,mr)
      integer, intent(in) :: ty
      logical, intent(in) :: cc(0:)
      integer, intent(out) :: mr
      select case(ty)
      case(4)
        if (cc(0) .or. cc(1)) then
          mr = 1
        else
          mr = 0
        end if 
      case(5)
        if (cc(0)) then
          mr = 1
        else
          mr = 0
        end if 
      case(7)
        if (cc(0) .or. cc(1) .or. cc(2)) then
          mr = 2
        else
          mr = 0
        end if 
      case(8)
        if (cc(0) .or. cc(1)) then
          mr = 3
        else
          mr = 0
        end if 
      case(9)
        if (cc(0) .or. cc(2) .or. cc(4)) then
          mr = 2
        else if(cc(1) .or. cc(3) .or. cc(5)) then
          mr = 1
        else
          mr = 0
        end if 
      case(11)
        if (cc(0) .or. cc(1) .or. cc(2)) then
          mr = 3
        else
          mr = 0
        end if 
      case(12)
        if (cc(0) .or. cc(1) .or. cc(2)) then
          mr = 3
        else
          mr = 0
        end if 
      case(13)
        if (cc(0) .or. cc(1) .or. cc(2)) then
          mr = 2
        else
          mr = 0
        end if 
      case(14)
        if (cc(0) .or. cc(1) .or. cc(2) .or. cc(3)) then
          mr = 3
        else
          mr = 0
        end if 
      case(15)
        if (cc(0) .or. cc(1) .or. cc(2)) then
          mr = 2
        else
          mr = 0
        end if 
      case(16)
        if (cc(0) .or. cc(1) .or. cc(2)) then
          mr = 1
        else
          mr = 0
        end if 
      case(17)
        if (cc(0) .or. cc(1) .or. cc(2)) then
          mr = 3
        else
          mr = 0
        end if 
      case(19)
        if (cc(0)) then
          mr = 1
        else
          mr = 0
        end if 
      case(25)
        if (cc(0) .or. cc(1) .or. cc(2) .or. cc(3)) then
          mr = 2
        else if(cc(4) .or. cc(5) .or. cc(6) .or. cc(7)) then
          mr = 1
        else
          mr = 0
        end if 
      case(26)
        if (cc(0)) then
          mr = 1
        else
          mr = 0
        end if 
      case(27)
        if (cc(0) .or. cc(1) .or. cc(2) .or. cc(3)) then
          mr = 2
        else
          mr = 0
        end if 
      case(28)
        if (cc(0) .or. cc(1) .or. cc(2) .or. cc(3)) then
          mr = 2
        else if(cc(5) .or. cc(7)) then
          mr = 1
        else
          mr = 0
        end if 
      case(29)
        if (cc(0) .or. cc(1) .or. cc(2) .or. cc(3)) then
          mr = 2
        else
          mr = 0
        end if 
      case(30)
        if (cc(0) .or. cc(1) .or. cc(2) .or. cc(3)) then
          mr = 2
        else if(cc(4) .or. cc(5)) then
          mr = 1
        else
          mr = 0
        end if 
      case(31)
        if (cc(0) .or. cc(1) .or. cc(2)) then
          mr = 2
        else
          mr = 0
        end if 
      case(32)
        if (cc(0) .or. cc(1) .or. cc(2) .or. cc(3)) then
          mr = 2
        else if(cc(4) .or. cc(5) .or. cc(6) .or. cc(7)) then
          mr = 1
        else
          mr = 0
        end if 
      case(33)
        if (cc(0) .or. cc(1) .or. cc(2) .or. cc(3)) then
          mr = 2
        else
          mr = 0
        end if 
      case(34)
        if (cc(0) .or. cc(1) .or. cc(2) .or. cc(3) .or. cc(4)) then
          mr = 2
        else
          mr = 0
        end if 
      case(35)
        if (cc(0) .or. cc(1) .or. cc(2) .or. cc(3)) then
          mr = 2
        else if(cc(4) .or. cc(5)) then
          mr = 1
        else
          mr = 0
        end if 
      case(36)
        if (cc(0) .or. cc(1) .or. cc(2)) then
          mr = 2
        else
          mr = 0
        end if 
      case(37)
        if (cc(0) .or. cc(1) .or. cc(2) .or. cc(3)) then
          mr = 2
        else
          mr = 0
        end if 
      case(38)
        if (cc(0) .or. cc(1) .or. cc(2)) then
          mr = 2
        else
          mr = 0
        end if 
      case(39)
        if (cc(0) .or. cc(1) .or. cc(2) .or. cc(3) .or. cc(4)) then
          mr = 2
        else
          mr = 0
        end if 
      case(40)
        if (cc(0) .or. cc(1) .or. cc(2) .or. cc(3)) then
          mr = 2
        else
          mr = 0
        end if 
      case(41)
        if (cc(0) .or. cc(1) .or. cc(2) .or. cc(3) .or. cc(4)) then
          mr = 2
        else
          mr = 0
        end if 
      case(42)
        if (cc(0) .or. cc(1) .or. cc(2) .or. cc(3)) then
          mr = 2
        else
          mr = 0
        end if 
      case(43)
        if (cc(0) .or. cc(1) .or. cc(2)) then
          mr = 2
        else
          mr = 0
        end if 
      case(44)
        if (cc(0) .or. cc(1) .or. cc(2) .or. cc(3)) then
          mr = 2
        else
          mr = 0
        end if 
      case(45)
        if (cc(0) .or. cc(1)) then
          mr = 1
        else
          mr = 0
        end if 
      case(46)
        if (cc(0)) then
          mr = 1
        else
          mr = 0
        end if 
      case(47)
        if (cc(0)) then
          mr = 1
        else
          mr = 0
        end if 
      case(49)
        if (cc(0) .or. cc(1)) then
          mr = 1
        else
          mr = 0
        end if 
      case(50)
        if (cc(0)) then
          mr = 1
        else
          mr = 0
        end if 
      case(51)
        if (cc(0)) then
          mr = 1
        else
          mr = 0
        end if 
      case(55)
        if (cc(0)) then
          mr = 1
        else
          mr = 0
        end if 
      case(56)
        if (cc(0) .or. cc(1)) then
          mr = 1
        else
          mr = 0
        end if 
      case(57)
        if (cc(0)) then
          mr = 1
        else
          mr = 0
        end if 
      case(69)
        if (cc(0)) then
          mr = 1
        else
          mr = 0
        end if 
      case(70)
        if (cc(0) .or. cc(1)) then
          mr = 1
        else
          mr = 0
        end if 
      case(71)
        if (cc(0)) then
          mr = 1
        else
          mr = 0
        end if 
      case(72)
        if (cc(0)) then
          mr = 1
        else
          mr = 0
        end if 
      case(73)
        if (cc(0) .or. cc(1)) then
          mr = 1
        else
          mr = 0
        end if 
      case(74)
        if (cc(0)) then
          mr = 1
        else
          mr = 0
        end if 
      case(75)
        if (cc(0)) then
          mr = 1
        else
          mr = 0
        end if 
      case(76)
        if (cc(0) .or. cc(1) .or. cc(2)) then
          mr = 3
        else
          mr = 0
        end if 
      case(77)
        if (cc(1) .or. cc(3) .or. cc(4)) then
          mr = 2
        else if(cc(0) .or. cc(2) .or. cc(5)) then
          mr = 1
        else
          mr = 0
        end if 
      case(79)
        if (cc(0) .or. cc(1) .or. cc(2)) then
          mr = 3
        else
          mr = 0
        end if 
      case(80)
        if (cc(0) .or. cc(1) .or. cc(2)) then
          mr = 2
        else
          mr = 0
        end if 
      case(81)
        if (cc(0) .or. cc(1) .or. cc(2) .or. cc(3)) then
          mr = 3
        else
          mr = 0
        end if 
      case(82)
        if (cc(0) .or. cc(1) .or. cc(2)) then
          mr = 3
        else
          mr = 0
        end if 
      case(83)
        if (cc(0) .or. cc(1)) then
          mr = 3
        else
          mr = 0
        end if 
      case(84)
        if (cc(0) .or. cc(1) .or. cc(2)) then
          mr = 1
        else
          mr = 0
        end if 
      case(85)
        if (cc(0) .or. cc(1)) then
          mr = 1
        else
          mr = 0
        end if 
      case(86)
        if (cc(0)) then
          mr = 1
        else
          mr = 0
        end if 
      case(87)
        if (cc(0)) then
          mr = 1
        else
          mr = 0
        end if 
      case(88)
        if (cc(0) .or. cc(1)) then
          mr = 2
        else
          mr = 0
        end if 
      case(89)
        if (cc(0)) then
          mr = 1
        else
          mr = 0
        end if 
      case(92)
        if (cc(0) .or. cc(1)) then
          mr = 2
        else
          mr = 0
        end if 
      case(93)
        if (cc(0)) then
          mr = 1
        else
          mr = 0
        end if 
      case(94)
        if (cc(0) .or. cc(1) .or. cc(2) .or. cc(3) .or. cc(4) .or. cc(5)) then
          mr = 1
        else
          mr = 0
        end if 
      case(96)
        if (cc(0) .or. cc(1) .or. cc(2)) then
          mr = 2
        else
          mr = 0
        end if 
      case(97)
        if (cc(0) .or. cc(1) .or. cc(2)) then
          mr = 2
        else
          mr = 0
        end if 
      case(98)
        if (cc(0) .or. cc(1)) then
          mr = 2
        else
          mr = 0
        end if 
      case(99)
        if (cc(0) .or. cc(1) .or. cc(2)) then
          mr = 1
        else
          mr = 0
        end if 
      case(100)
        if (cc(0) .or. cc(1) .or. cc(2) .or. cc(3)) then
          mr = 2
        else
          mr = 0
        end if 
      case(101)
        if (cc(0) .or. cc(1) .or. cc(2)) then
          mr = 1
        else
          mr = 0
        end if 
      case(105)
        if (cc(0) .or. cc(1)) then
          mr = 2
        else
          mr = 0
        end if 
      case(106)
        if (cc(0)) then
          mr = 1
        else
          mr = 0
        end if 
      case(108)
        if (cc(0)) then
          mr = 1
        else
          mr = 0
        end if 
      case(111)
        if (cc(0) .or. cc(1)) then
          mr = 1
        else
          mr = 0
        end if 
      case(112)
        if (cc(0)) then
          mr = 1
        else
          mr = 0
        end if 
      case(114)
        if (cc(0)) then
          mr = 1
        else
          mr = 0
        end if 
      case default
        mr = 0
      end select 
    end subroutine get_vertex_lorentz_rank_mdl
end module class_vertices
